#include "..\config.h"

#ifdef INCLUDE_REBUILD

namespace bnf {

#include "main.h"

#define DEBUG 0

struct SymbolStruct
{
	char* SymbolName;
	int* SymbolGrammar;
	int SymbolGrammarSize;
	char* SymbolPointer;
	int Compressed;
};

struct SymbolReferenceStruct
{
	char* SymbolName;
};

SymbolStruct* Symbols;
int SymbolCount = 0;
int CurSymbol = 0;
SymbolReferenceStruct* SymbolReferences;
int SymbolReferenceCount = 0;
int CurSymbolReference = 0;

char* SymbolBuffer;

int SymbolTreeBufferSize;

char* GrammarTree;
int GrammarTreeSize;

char* pInitSymbol;

bool IdentifierChar(char character)
{
	return((character >= '0' && character <= '9') ||
		((character | 32) >= 'a' && (character | 32) <= 'z') ||
		(character == '_'));
}

char* ProcessIdentifier(char* GrammarPos)
{
	char* BufferPos = SymbolBuffer;
	while (IdentifierChar(*GrammarPos))
	{
		*(BufferPos++) = *(GrammarPos++);
	}
	*BufferPos = 0;
	if (DEBUG) {printf("Identifier Name: %s\n", SymbolBuffer);}
	return(GrammarPos);
}

char* ProcessString(char* GrammarPos, char EndChar)
{
	char* BufferPos = SymbolBuffer;
	while (*GrammarPos)
	{
		if (*GrammarPos++ == EndChar)
		{
			*BufferPos = 0;
			if (DEBUG) {printf("String: %s\n", SymbolBuffer);}
			return(GrammarPos);
		}
		else
		{
			*(BufferPos++) = *(GrammarPos - 1);
		}
	}

	return(0);
}

template <class T> void AddSymbolTreeBuffer(char* SymbolTreeBuffer, int* SymbolTreeSize, T val)
{
	 *((int*) (SymbolTreeBuffer + *SymbolTreeSize)) = val;
	 *SymbolTreeSize += sizeof(val);
}

int AddReference(char* SymbolName)
{
	if (!(SymbolReferences[CurSymbolReference].SymbolName = (char*) mallocl(strlen(SymbolName) + 1, "AddReference-SymbolName")))
	{
		printf("Memory Allocation Error\n");
		return(-1);
	}
	strcpy(SymbolReferences[CurSymbolReference].SymbolName, SymbolName);
	return(CurSymbolReference++);
}

char* ProcessGrammarPart(char* GrammarPos, char EndChar, char* SymbolName)
{
	int i;

	char* SymbolTreeBuffer;
	int SymbolTreeSize = 0;

	char ASCBuffer[4];

	int ReferenceID;

	if (DEBUG) {printf("Rule: %s\n", SymbolName);}

	int CurSymbol = bnf::CurSymbol;
	bnf::CurSymbol++;
	if ((!(Symbols[CurSymbol].SymbolName = (char*) mallocl(strlen(SymbolName) + 1, "ProcessGrammarPart-Symbols-Name"))) ||
		(!(SymbolTreeBuffer = (char*) mallocl(SymbolTreeBufferSize, "ProcessGrammarPart-SymbolTreeBuffer"))))
	{
		printf("Memory Allocation Error\n");
		return(0);
	}
	strcpy(Symbols[CurSymbol].SymbolName, SymbolName);
	SymbolTreeSize = 0;

	int SubSymbol = 0;
	char* SubSymbolName;
	SubSymbolName = (char*) mallocl(strlen(SymbolName) + 16, "ProcessGrammarPart-SubSymbolName");

	while (GrammarPos != 0 && *GrammarPos)
	{
		if (*GrammarPos == '"')
		{
			if (DEBUG) {printf("Found String\n");}
			GrammarPos = ProcessString(++GrammarPos, '"');
			AddSymbolTreeBuffer<int>(SymbolTreeBuffer, &SymbolTreeSize, G_STRING);
			AddSymbolTreeBuffer<int>(SymbolTreeBuffer, &SymbolTreeSize, strlen(SymbolBuffer));
			CopyMemory(SymbolTreeBuffer + SymbolTreeSize, SymbolBuffer, strlen(SymbolBuffer));
			SymbolTreeSize += strlen(SymbolBuffer);
		}
		else if (*GrammarPos == '\'')
		{
			if (DEBUG) {printf("Found String\n");}
			GrammarPos = ProcessString(++GrammarPos, '\'');
			AddSymbolTreeBuffer<int>(SymbolTreeBuffer, &SymbolTreeSize, G_STRING);
			AddSymbolTreeBuffer<int>(SymbolTreeBuffer, &SymbolTreeSize, strlen(SymbolBuffer));
			CopyMemory(SymbolTreeBuffer + SymbolTreeSize, SymbolBuffer, strlen(SymbolBuffer));
			SymbolTreeSize += strlen(SymbolBuffer);
		}
		else if (*GrammarPos == '(')
		{
			if (DEBUG) {printf("Found Group\n");}
			sprintf(SubSymbolName, "%s sub %d", SymbolName, SubSymbol);
			GrammarPos = ProcessGrammarPart(++GrammarPos, ')', SubSymbolName);
			AddSymbolTreeBuffer<int>(SymbolTreeBuffer, &SymbolTreeSize, G_GROUP);
			ReferenceID = AddReference(SubSymbolName);
			if (ReferenceID == -1) {return(0);}
			AddSymbolTreeBuffer<int>(SymbolTreeBuffer, &SymbolTreeSize, ReferenceID);
			SubSymbol++;
		}
		else if (*GrammarPos == '{')
		{
			if (DEBUG) {printf("Found Rep\n");}
			sprintf(SubSymbolName, "%s sub %d", SymbolName, SubSymbol);
			GrammarPos = ProcessGrammarPart(++GrammarPos, '}', SubSymbolName);
			AddSymbolTreeBuffer<int>(SymbolTreeBuffer, &SymbolTreeSize, G_REP);
			ReferenceID = AddReference(SubSymbolName);
			if (ReferenceID == -1) {return(0);}
			AddSymbolTreeBuffer<int>(SymbolTreeBuffer, &SymbolTreeSize, ReferenceID);
			SubSymbol++;
		}
		else if (*GrammarPos == '[')
		{
			if (DEBUG) {printf("Found Opt\n");}
			sprintf(SubSymbolName, "%s sub %d", SymbolName, SubSymbol);
			GrammarPos = ProcessGrammarPart(++GrammarPos, ']', SubSymbolName);
			AddSymbolTreeBuffer<int>(SymbolTreeBuffer, &SymbolTreeSize, G_OPT);
			ReferenceID = AddReference(SubSymbolName);
			if (ReferenceID == -1) {return(0);}
			AddSymbolTreeBuffer<int>(SymbolTreeBuffer, &SymbolTreeSize, ReferenceID);
			SubSymbol++;
		}
		else if (*GrammarPos == ',')
		{
			if (DEBUG) {printf("Found Seperator\n");}
			AddSymbolTreeBuffer<int>(SymbolTreeBuffer, &SymbolTreeSize, G_SEP);
			GrammarPos++;
		}
		else if (*GrammarPos >= '0' && *GrammarPos <= '9')
		{
			*((int*) ASCBuffer) = 0;
			i = 0;
			while (*GrammarPos >= '0' && *GrammarPos <= '9')
			{
				ASCBuffer[i] = *(GrammarPos++);
				i++;
			}
			i = atoi(ASCBuffer);
			AddSymbolTreeBuffer<int>(SymbolTreeBuffer, &SymbolTreeSize, G_STRING);
			AddSymbolTreeBuffer<int>(SymbolTreeBuffer, &SymbolTreeSize, 1);
			*(SymbolTreeBuffer + SymbolTreeSize) = (char) i;
			SymbolTreeSize ++;
		}
		else if (IdentifierChar(*GrammarPos))
		{
			if (DEBUG) {printf("Found Identifier\n");}
			GrammarPos = ProcessIdentifier(GrammarPos);
			AddSymbolTreeBuffer<int>(SymbolTreeBuffer, &SymbolTreeSize, G_LINK);
			ReferenceID = AddReference(SymbolBuffer);
			if (ReferenceID == -1) {return(0);}
			AddSymbolTreeBuffer<int>(SymbolTreeBuffer, &SymbolTreeSize, ReferenceID);
		}
		else if (*GrammarPos == '|')
		{
			if (DEBUG) {printf("Found \"|\"\n");}
			AddSymbolTreeBuffer<int>(SymbolTreeBuffer, &SymbolTreeSize, G_OR);
			GrammarPos++;
		}
		else if (*GrammarPos == '-')
		{
			if (DEBUG) {printf("Found \"-\"\n");}
			AddSymbolTreeBuffer<int>(SymbolTreeBuffer, &SymbolTreeSize, G_NOT);
			GrammarPos++;
		}
		else if (*GrammarPos++ == EndChar)
		{
			if (DEBUG) {printf("Found Rule End\n");}
			AddSymbolTreeBuffer<int>(SymbolTreeBuffer, &SymbolTreeSize, G_END);
			if (!(Symbols[CurSymbol].SymbolGrammar = (int*) mallocl(SymbolTreeSize, "ProcessGrammarPart-Symbols-Grammar")))
			{
				printf("Memory Allocation Error\n");
				return(0);
			}

			CopyMemory(Symbols[CurSymbol].SymbolGrammar, SymbolTreeBuffer, SymbolTreeSize);
			Symbols[CurSymbol].SymbolGrammarSize = SymbolTreeSize;

			freel(SubSymbolName);
			freel(SymbolTreeBuffer);
				
			return(GrammarPos);
		}
		else
		{
			return(0);
		}
	}
	return(0);
}

char* ProcessRule(char* GrammarPos)
{
	char* BufferPos = SymbolBuffer;
	char* SymbolName;
	char* retVal;

	while (*GrammarPos)
	{
		if (IdentifierChar(*GrammarPos))
		{
			*(BufferPos++) = *(GrammarPos++);
		}
		else if (*GrammarPos == '=')
		{
			*BufferPos = 0;
			SymbolName = (char*) mallocl(strlen(SymbolBuffer) + 1, "ProcessRule-SymbolName");

			strcpy(SymbolName, SymbolBuffer);
			retVal = ProcessGrammarPart(++GrammarPos, ';', SymbolName);
			freel(SymbolName);
			return(retVal);
		}
		else
		{
			return(0);
		}
	}
	return(0);
}

int ResolveSymbolAddress(int* Reference)
{
	int i;

	if (strcmp(SymbolReferences[*Reference].SymbolName, "anycharacter") == 0)
	{
		*Reference = (int) 0;
		return(0);
	}

	for (i = 0;i < CurSymbol;i++)
	{
		if (strcmp(Symbols[i].SymbolName, SymbolReferences[*Reference].SymbolName) == 0)
		{
			*Reference = (int) Symbols[i].SymbolPointer;
			return(0);
		}
	}
	printf("Internal Error\n");
	return(1);
}

void StripGrammar(char* GrammarBuffer, int* GrammarSize)
{
	int i, j;
	int Str1 = 0, Str2 = 0;

	if (DebugOutput)
	{
		printf("Striping Grammar\n");
	}
	j = 0;
	for (i = 0;i < *GrammarSize;i++)
	{
		if (GrammarBuffer[i] == '(' || GrammarBuffer[i] == '{' || GrammarBuffer[i] == '[' || GrammarBuffer[i] == '=')
		{
			SymbolCount++;
		}
		if (GrammarBuffer[i] == '(' || GrammarBuffer[i] == '{' || GrammarBuffer[i] == '[' || GrammarBuffer[i] == '=' || GrammarBuffer[i] == ',' || GrammarBuffer[i] == '|' || GrammarBuffer[i] == '-')
		{
			SymbolReferenceCount++;
		}
		if ((GrammarBuffer[i] != ' ' && GrammarBuffer[i] != '	' && GrammarBuffer[i] != 0 && GrammarBuffer[i] != 10 && GrammarBuffer[i] != 13) ||
			Str1 || Str2)
		{
			if (GrammarBuffer[i] == '"' && Str2 == 0)
			{
				Str1 ^= 1;
			}

			if (GrammarBuffer[i] == '\'' && Str1 == 0)
			{
				Str2 ^= 1;
			}

			if (i != j)
			{
				GrammarBuffer[j] = GrammarBuffer[i];
			}
			j++;
		}
	}
	*GrammarSize = j;
	SymbolTreeBufferSize = *GrammarSize * 2 * sizeof(int);
	GrammarBuffer[j] = 0;
}

int MarkCompressedSymbols()
{
	char* Buffer;
	char* BufferPos;
	int i, j, k;

	if (DebugOutput)
	{
		printf("Marking Compressed Symbols\n");
	}

	if (GetPrivateProfileStringA("Config", "CompressedSymbol", &Buffer))
	{
		return(1);
	}

	for (i = 0;i < CurSymbol;i++)
	{
		Symbols[i].Compressed = 0;
	}

	j = 0;
	k = strlen(Buffer);
	for (i = 0;i < k;i++)
	{
		if (Buffer[i] != ' ' && Buffer[i] != '	')
		{
			if (Buffer[i] == ',')
			{
				Buffer[j] = 0;
			}
			else
			{
				Buffer[j] = Buffer[i];
			}
			j++;
		}
	}
	if (Buffer[j - 1] != 0)
	{
		Buffer[j] = 0;
		j++;
	}
	Buffer[j] = 0;

	BufferPos = Buffer;
	while (*BufferPos)
	{
		for (i = 0;i < CurSymbol;i++)
		{
			if (strcmp(Symbols[i].SymbolName, BufferPos) == 0)
			{
				Symbols[i].Compressed = 1;
				goto CompressedSymbolFound;
			}
		}
		printf("Could not find Symbol %s\n", BufferPos);
		return(1);

CompressedSymbolFound:
		BufferPos += strlen(BufferPos) + 1;
	}

	freel(Buffer);

	return(0);
}

int ProcessGrammar(char* GrammarBuffer, int GrammarSize)
{
	int i, j;
	char* GrammarPos;
	char* GrammarTreePos;
	int SubSymbolCount;
	bool PrintSymbolName;
	int CurOrder;
	char* InitSymbol;
	char* pSymbolName;

	StripGrammar(GrammarBuffer, &GrammarSize);
	GrammarPos = GrammarBuffer;

	if (DebugOutput)
	{
		printf("Allocating Symbols (%d / %d bytes)\n", sizeof(SymbolStruct) * SymbolCount, sizeof(SymbolReferenceStruct) * SymbolReferenceCount);
	}
	Symbols = (SymbolStruct*) mallocl(sizeof(SymbolStruct) * SymbolCount, "ProcessGrammar-Symbols");
	SymbolReferences = (SymbolReferenceStruct*) mallocl(sizeof(SymbolReferenceStruct) * SymbolReferenceCount, "ProcessGrammar-SymbolReferences");

	j = 0;
	do
	{
		if (DEBUG) {printf("Reading Rule %d\n", j);}
		GrammarPos = ProcessRule(GrammarPos);
		if (GrammarPos == 0)
		{
			printf("Error while processing Grammar\n");
			return(1);
		}
		j++;
	}
	while (*GrammarPos != 0);

	if (DebugOutput)
	{
		printf("Checking References\n");
	}
	for (i = 0;i < CurSymbolReference;i++)
	{
		for (j = 0;j < CurSymbol;j++)
		{
			if (strcmp(SymbolReferences[i].SymbolName, "anycharacter") == 0)
			{
				goto SymbolFound;
			}
			else if (strcmp(SymbolReferences[i].SymbolName, Symbols[j].SymbolName) == 0)
			{
				goto SymbolFound;
			}
		}
		printf("Coult not find Symbol %s\n", SymbolReferences[i].SymbolName);
		return(1);

SymbolFound:;
	}

	if (GetPrivateProfileStringA("Config", "InitSymbol", &InitSymbol))
	{
		return(1);
	}

	if (InitSymbol[0] == 0)
	{
		printf("Could not read InitSymbol from Config File\n");
		freel(InitSymbol);
		return(1);
	}

	for (j = 0;j < CurSymbol;j++)
	{
		if (strcmp(Symbols[j].SymbolName, InitSymbol) == 0)
		{
			pInitSymbol = (char*) &Symbols[j];
			goto InitSymbolFound;
		}
	}

	freel(InitSymbol);

	printf("Could not find Init Rule\n");
	return(1);

InitSymbolFound:
	freel(InitSymbol);
	if (DebugOutput)
	{
		printf("All References Present (%d)\n", CurSymbolReference);
	}

	GrammarTreeSize = 0;
	SubSymbolCount = 0;
	SymbolNameSize = 0;
	for (i = 0;i < CurSymbol;i++)
	{
		GrammarTreeSize += Symbols[i].SymbolGrammarSize;
		SymbolNameSize += strlen(Symbols[i].SymbolName) + 1;
		if (strstr(Symbols[i].SymbolName, " ") != 0) {SubSymbolCount++;}
	}
	GrammarTreeSize += CurSymbol * sizeof(int) * 2;
	GrammarTreeSize += SymbolNameSize;

	if (DebugOutput)
	{
		printf("Size of Grammar: %d (%d Symbols, %d SubSymbols)\n", GrammarTreeSize, CurSymbol, SubSymbolCount);
	}
	if (!(GrammarTree = (char*) mallocl(GrammarTreeSize, "ProcessGrammar-GrammarTree")))
	{
		printf("Memory Allocation Error\n");
		return(1);
	}

	if (MarkCompressedSymbols()) {return(1);}

	if (DebugOutput)
	{
		printf("Intermixing Symbols\n");
	}
	GrammarTreePos = GrammarTree;
	for (i = 0;i < CurSymbol;i++)
	{
		Symbols[i].SymbolPointer = GrammarTreePos;
		*((int*) GrammarTreePos) = Symbols[i].Compressed;
		GrammarTreePos += sizeof(int);
		*((int*) GrammarTreePos) = i;
		GrammarTreePos += sizeof(int);
		strcpy(GrammarTreePos, Symbols[i].SymbolName);
		GrammarTreePos += strlen(Symbols[i].SymbolName) + 1;
		CopyMemory(GrammarTreePos, Symbols[i].SymbolGrammar, Symbols[i].SymbolGrammarSize);
		GrammarTreePos += Symbols[i].SymbolGrammarSize;
	}

	if (DebugOutput)
	{
		printf("Resolving References\n");
	}

	pInitSymbol = (char*) ((SymbolStruct*) pInitSymbol)->SymbolPointer;

	GrammarTreePos = GrammarTree;
	PrintSymbolName = true;
	while (GrammarTreePos < GrammarTree + GrammarTreeSize)
	{
		if (PrintSymbolName)
		{
			GrammarTreePos += 8;
			if (DEBUG){printf("Done with Symbol %s\n", GrammarTreePos);}
			GrammarTreePos += strlen(GrammarTreePos) + 1;
			PrintSymbolName = false;
		}
		
		CurOrder = *((int*) GrammarTreePos);
		
		if (CurOrder == G_END)
		{
			PrintSymbolName = true;
		}
		
		if (CurOrder == G_GROUP || CurOrder == G_REP || CurOrder == G_OPT || CurOrder == G_LINK) 
		{
			if (ResolveSymbolAddress((int*) (GrammarTreePos + 4)))
			{
				return(1);
			}
		}

		GrammarTreePos = SkipRule(GrammarTreePos);
	}

	if (DebugOutput)
	{
		printf("Creating Symbol List\n");
	}
	SymbolNameSizeCopy = SymbolNameSize + 1;
	SymbolNames = (char*) mallocl(SymbolNameSizeCopy, "ProcessGrammar-SymbolNames");
	SymbolPointer = (char**) mallocl(CurSymbol * sizeof(int), "ProcessGrammar-SymbolPointer");

	if (SymbolNames == 0 || SymbolPointer == 0)
	{
		printf("Memory Allocation Error\n");
		return(1);
	}

	pSymbolName = SymbolNames;
	for (i = 0;i < CurSymbol;i++)
	{
		SymbolPointer[i] = (char*) (pSymbolName - SymbolNames);
		j = strlen(Symbols[i].SymbolName);
		CopyMemory(pSymbolName, Symbols[i].SymbolName, j);
		pSymbolName[j] = 0;
		pSymbolName += ++j;
	}
	SymbolNames[SymbolNameSize] = 0;

	if (DebugOutput)
	{
		printf("Freeing Temporary Symbols and References\n");
	}
	for (i = 0;i < CurSymbol;i++)
	{
		freel(Symbols[i].SymbolName);
		freel(Symbols[i].SymbolGrammar);
	}
	for (i = 0;i < CurSymbolReference;i++)
	{
		freel(SymbolReferences[i].SymbolName);
	}
	freel(Symbols);
	freel(SymbolReferences);

	return(0);
}

int CreateGrammar()
{
	unsigned long i;
	unsigned long GrammarSize;
	char* GrammarBuffer = NULL;

	char* GrammarFile;

	if (GetPrivateProfileStringA("Config", "GrammarFile", &GrammarFile))
	{
		return(1);
	}

	if (DebugOutput)
	{
		printf("Opening Grammar File\n");
	}
	if (!(GrammarSize = ReadFileToBuffer(GrammarFile, &GrammarBuffer)))
	{
		printf("Error loading Grammar File\n");
		return(1);
	}

	if (DebugOutput)
	{
		printf("Allocation Symbol Buffer\n");
	}
	SymbolBuffer = (char*) mallocl(GrammarSize, "CreateGrammar-SymbolBuffer");
	if (SymbolBuffer == 0)
	{
		printf("Memory Allocation Error\n");
		return(1);
	}

	if (DebugOutput)
	{
		printf("Processing Grammar\n");
	}
	i = ProcessGrammar(GrammarBuffer, GrammarSize);

	if (DebugOutput)
	{
		printf("Freeing Memory\n");
	}
	freel(GrammarBuffer);
	freel(SymbolBuffer);
	freel(GrammarFile);

	return(i);
}

int InitGrammar()
{
	if (DebugOutput)
	{
		printf("Initializing Grammar\n");
	}
	if (CreateGrammar() == 0)
	{
		if (DebugOutput)
		{
			printf("Finished\n");
		}
		return(0);
	}
	else
	{
		printf("Error\n");
		return(1);
	}
}

int ExitGrammar()
{
	printf("Unloading Grammar\n");
	printf("Freeing Grammar Tree\n");
	freel(GrammarTree);

	return(0);
}

}

#endif